(require
  hyrule [unless]
  kodhy.macros [$ getl ss ssi wc cbind wcby])

(import
  json
  re
  datetime [date]
  numpy :as np
  pandas :as pd
  seaborn :as sns
  scikits.bootstrap
  metadict [MetaDict]
  toolz [unique first last]
  kodhy.util [F T ucfirst valcounts rd])

(setv rating-scales (dfor
  [k v] (.items (dict
    :satisfaction #("Very satisfied" "Mostly satisfied" "Somewhat satisfied" "Mostly unsatisfied" "Not at all satisfied")
    :quality #("A very good decision" "A good decision" "A so-so decision" "A bad decision" "A very bad decision")
    :easiness #("Very easy" "Easy" "Intermediate" "Hard" "Very hard")
    :vda_consistency #("Very similar" "Mostly similar" "Somewhat similar" "Mostly dissimilar" "Not at all similar")
    :vda_easiness #("Very easy" "Easy" "Intermediate" "Hard" "Very hard")
    :vda_helpfulness #("Very helpful" "Mostly helpful" "Somewhat helpful" "Mostly unhelpful" "Not at all helpful")))
  k (dict (zip [1 2 3 4 5] (reversed v)))))

(defn jload [x] (with [o (open f"data/{x}.json")] (json.load o)))
(setv
  data (jload "main")
  visit1-invited (jload "visit1_invited")
  desc-categories (jload "desc_categories")
  ratings (jload "ratings"))

(defn intify-keys [d]
  (setv new (tuple (zip (map int (.keys d)) (.values d))))
  (.clear d)
  (.update d new))

(intify-keys data)
(for [subject data]
  (intify-keys (get data subject "sessions")))
(intify-keys (get desc-categories "subjects"))

(for [[k v] (list (.items ratings))]
  (intify-keys v)
  (setv (get ratings k) (. (pd.DataFrame v) T)))
(assert (=
  (list (. ratings ["pilot"] columns))
  ["reversed_criterion" "opaque" "desc_mismatch"]))
(assert (=
  (list (. ratings ["round 1, rerating (Kodi B. Arfer)"] columns))
  ["reversed_criterion" "bad_alt_level" "other_issue"]))
(assert (=
  (list (. ratings ["round 1, rerating (Kelly Matula)"] columns))
  ["reversed_criterion" "bad_alt_level"]))
; I spoke to Kelly by email on 20 Mar 2023 to reconcile our codes in
; the 3 cases we'd disagreed, and we agreed to go with my codes.
(setv (get ratings "round1")
  (get ratings "round 1, rerating (Kodi B. Arfer)"))

(setv task-versions (list (unique (gfor
  [_ commit] (sorted (gfor
    subject (.values data)
    session (.values (get subject "sessions"))
    #(
      (or
        (* (.get session "accepted" 0) 1000)
        (.get session "time_started")
        (get session "time_started_posixms"))
      (get session "task_version"))))
  commit))))

(defn posixms->date [x]
  (.floor (pd.Timestamp :unit "ms" x) "D"))

(setv subjects (.set-index :keys "sn" (pd.DataFrame (gfor
  [sn d] (.items data)
  ; Only use the Prolific sample.
  :if (= (get d "recruited") "prolific")
  (dict
    :sn (int sn)
    :prolific-pid (.get d "prolific_pid")
    :began (posixms->date (get d "sessions" 0 "time_started_posixms"))
    :first-task-commit (get (first (.values (get d "sessions"))) "task_version")
    :tv 0
    :last-visit (max (.keys (get d "sessions")))
    :last-task-commit (get (last (.values (get d "sessions"))) "task_version")
    :last-tv 0
    :cond (get d "cond")
    #** (if (not-in 1 (get d "sessions")) {} (dict
      :expected-resolution-date (do
        (setv x (get d "sessions" 1 "expected_resolution_date"))
        (if (and
            (setx m (re.match r"(\d\d\d\d)-\d\d-\d\d\Z" x))
            (>= (int (.group m 1)) 2000))
          (pd.to-datetime x)
          None))))
    #** (if (not-in "time_puzzle" (.get (get d "sessions") 1 {})) {} (dict
      :time-puzzle-minutes (/
        (get d "sessions" 1 "time_puzzle")
        (* 60 1000))
      :n-puzzle-attempts
        (len (get d "sessions" 1 "puzzle_responses"))))
    #** (if (not-in 2 (get d "sessions")) {} (do
      (setv d2 (MetaDict (get d "sessions" 2)))
      (dict
        :visit2_date (posixms->date d2.time-started-posixms)
        :gender (if (in d2.gender ["male" "female"])
          (ucfirst d2.gender)
          "Other")
        :age d2.age
        :education_years (.get d2 "education_years" None)
        :country d2.country
        #** (dfor
          [k v] d2.race
          (+ "race_" k) v)
        :time_visit2_minutes (/
          d2.time-done
          (* 60 1000))
        :visit2_artiruno_commit d2.artiruno-commit)))
    :followed_up (in 3 (get d "sessions"))
    #** (if (not-in 3 (get d "sessions")) {} (do
      (setv d3 (MetaDict (get d "sessions" 3)))
      (dict
        :visit3_date (posixms->date d3.time-started-posixms)
        #** (dfor k ["eval_rate_easiness" "eval_rate_quality" "eval_rate_satisfaction" "eval_rate_vda_consistency" "eval_rate_vda_easiness" "eval_rate_vda_helpfulness"]
          k (.get d3 k))))))))))

(setv ($ subjects tv) (lfor
  x ($ subjects first_task_commit)
  (.index task-versions x)))
(setv ($ subjects last_tv) (lfor
  x ($ subjects last_task_commit)
  (.index task-versions x)))

(setv (getl subjects 132 "country") "gb")
  ; A correction per the subject's comments.

(setv ($ subjects bad_vda) (.isin subjects.index
  ; See the explanation in the notebook.
  (ssi (get ratings "round1") (|
    $reversed_criterion $bad_alt_level $other_issue))))

(setv ($ subjects round1) (wc subjects (&
  (.isin $tv [11 12])
  (pd.notnull $time_visit2_minutes))))

(setv vda-props (.set-index :keys "sn" (pd.DataFrame (gfor
  ; Various automatically detectable properties of the VDA setup and
  ; Artiruno's behavior.
  [sn d] (.items data)
  :setv vda (.get (get d "sessions") 2 {})
  :if (in "dominance" vda)
  :do (setv
    criteria (lfor  [_ levels] (get vda "criteria")  levels)
    alts (lfor [_ values] (get vda "alts")  values)
    questions (.get vda "vda_questions" []))
  (dict
    :sn sn
    :dominator (get vda "dominance")
      ; Is there one alt that dominates all the others?
    :dominated (any (gfor
      ; Is there an alt that's dominated by at least one other alt?
      ; (By construction of the task, we know that all alts are
      ; distinct.)
      a1 alts
      a2 alts
      :if (!= a1 a2)
      (all (gfor
        [ci [v1 v2]] (enumerate (zip a1 a2))
        :setv levels (get criteria ci)
        (<= (.index levels v1) (.index levels v2))))))
    :unused-level (any (gfor
      ; Is there a level that's never used in any of the alts? (It
      ; might still have appeared in a VDA question, if it's the best
      ; level for its criterion.)
      [ci levels] (enumerate criteria)
      level levels
      (all (lfor
        alt alts
        (!= (get alt ci) level)))))
    :constant-criterion (any (gfor
      ; Is there a criterion that's constant across all alts?
      [ci _] (enumerate criteria)
      (= 1 (len (sfor
        alt alts
        (get alt ci))))))
    :n-questions (len questions)
    :n-criteria (len criteria)
    :vda-varied-all-cs (any (gfor
      ; Did VDA ask a question in which all criteria were varied?
      q questions
      (all (gfor
        [ci _] (enumerate criteria)
        (!= (get q "a" ci) (get q "b" ci))))))
    :vda-redundant (any (gfor
      ; Did VDA straight-up ask the subject to decide between alts?
      q questions
      (and (in (get q "a") alts) (in (get q "b") alts))))
    :result-quality (cond
      (or (get vda "dominance") (.startswith (get vda "vda_result") "Your choices imply a single best alternative: "))
        "winner"
      (.startswith (get vda "vda_result") "Your choices imply that these alternatives are tied for the best: ")
        "tie"
      (= (get vda "vda_result") "Your choices imply that all alternatives are tied for the best.")
        "dodo"
      True
        (raise ValueError)))))))

(defn show-vda-setups [sns]
  (for [sn sns]
    (unless (and
        (= (get data sn "cond") "vda")
        (in 2 (get data sn "sessions")))
      (continue))
    (setv vda (MetaDict (get data sn "sessions" 2)))
    (print "SN" sn)
    (print (round (/ vda.time-done (* 1,000 60))) "minutes")
    (print "Scenario:" (get data sn "sessions" 1 "problem_description"))
    (print "Criteria:")
    (for [[name values] vda.criteria]
      (print "•" (+ name ":") #* (map hy.repr values)))
    (print "Alts:")
    (for [[name values] vda.alts]
      (print "•" (+ name "::") (.join " - " (gfor
        [cval [cname _]] (zip values vda.criteria)
        (+ cname ": " (hy.repr cval))))))
    (print)))

(defn comments-dt [sns]
  (.set-index :keys ["sn" "visit"] :drop T (pd.DataFrame (gfor
    sn sns
    [visit session] (.items (get data sn "sessions"))
    :if (setx comments (.get session "comments"))
    (dict
      :sn sn
      :visit visit
      :date (.date (pd.Timestamp :tz "UTC" :unit "ms"
        (get session "time_started_posixms")))
      :comments (.replace comments "\n" " • "))))))
